// Copyright 2024 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package jobqueue

import (
	"gorm.io/gorm"
	"strings"
	"time"
)

func (s *JobSyncer) DeleteJob(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	return s.manager.database.Transaction(func(tx *gorm.DB) error {
		permJob := job.GetPersistence()
		if err := tx.Where("job_id = ?", permJob.GetID()).Delete(&JobLog{}).Error; err != nil {
			return err
		}
		if err := tx.Where("job_id = ?", permJob.GetID()).Delete(&JobStats{}).Error; err != nil {
			return err
		}
		if err := tx.Delete(&permJob).Error; err != nil {
			return err
		}
		return nil
	})
}

func (s *JobSyncer) ResetLogs(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	return s.manager.database.Transaction(func(tx *gorm.DB) error {
		permJob := job.GetPersistence()
		if err := tx.Unscoped().Where("job_id = ?", permJob.GetID()).Delete(&JobLog{}).Error; err != nil {
			return err
		}

		return nil
	})
}

func (s *JobSyncer) ResetStats(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	job.ResetStats()
	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}
	return saveJob(job, s.manager.database)
	//stats := job.GetStats()
	//return s.manager.database.Transaction(func(tx *gorm.DB) error {
	//	return tx.Model(&JobStats{}).Where("job_id = ?",
	//		job.GetID()).
	//		Select("*").Omit("deleted_at", "created_at", "job_id").Updates(stats).Error
	//
	//})
}

func (s *JobSyncer) SaveJob(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()
	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}
	return saveJob(job, s.manager.database)
}

func save(job *JobPersistence, db *gorm.DB) error {

	if db == nil {
		return ErrNoDatabaseConnection
	}

	return db.Transaction(func(tx *gorm.DB) error {

		if err := tx.Save(job).Error; err != nil {
			return err
		}

		if job.Stats != (JobStats{}) {
			job.Stats.JobID = job.ID
			if err := tx.Save(&job.Stats).Error; err != nil {
				return err
			}
		}

		for i, _ := range job.Logs {
			job.Logs[i].LogID = 0
			_ = tx.Create(&job.Logs[i])
			// no error handling, if it fails, it fails
		}

		return nil
	})
}

func saveJob(job GenericJob, db *gorm.DB) error {

	if db == nil {
		return ErrNoDatabaseConnection
	}

	permJob := job.GetPersistence()

	maxRetries := 3
	var attempt int
	for attempt = 0; attempt < maxRetries; attempt++ {
		err := save(&permJob, db)
		if err != nil {
			if strings.Contains(err.Error(), "lock") {
				time.Sleep(time.Millisecond * 100)
				continue
			}
			return err
		}
		break
	}

	if attempt == maxRetries {
		return ErrMaxRetriesReached
	}

	return nil

	//return db.Transaction(func(tx *gorm.DB) error {
	//
	//	permJob := job.GetPersistence()
	//
	//	memLogs := permJob.Logs
	//	permJob.Logs = nil
	//
	//	var existingJob JobPersistence
	//	result := tx.Unscoped().Where("id = ?", permJob.GetID()).First(&existingJob)
	//
	//	if result.Error != nil {
	//		err, done := createJob(tx, result, permJob)
	//		if done {
	//			return err
	//		}
	//	} else {
	//
	//		err, done := updateJob(tx, existingJob, permJob)
	//		if done {
	//			return err
	//		}
	//	}
	//
	//	if tx.Error != nil {
	//		Trace("Error while updating job", "error", tx.Error)
	//		return tx.Error
	//	}
	//
	//	return nil
	//
	//})

}

//
//func updateJob(tx *gorm.DB, existingJob JobPersistence, permJob JobPersistence) (error, bool) {
//
//	r := tx.Unscoped().Model(&existingJob).Select("*").
//		Omit("deleted_at", "created_at", "job_id").
//		Update("deleted_at", nil)
//
//	if r.Error != nil {
//		Trace("Error while deleting job", "error", r.Error)
//		return r.Error, true
//	}
//
//	resultStats := tx.Model(&existingJob).Select("*").
//		Omit("deleted_at", "created_at", "job_id").Update()
//
//	if resultStats.Error != nil {
//		Trace("Error while updating job stats", "error", resultStats.Error)
//		return resultStats.Error, true
//	}
//
//	r2 := tx.Model(&existingJob).Select("*").
//		Omit("deleted_at", "created_at", "job_id").Updates(permJob)
//	if r2.Error != nil {
//		return r2.Error, true
//	}
//
//	tx.Model(&permJob.Stats).
//		Select("*").Omit("deleted_at", "created_at", "job_id").
//		UpdateColumns(permJob.Stats)
//
//	if tx.Error != nil {
//		Error("Error while updating job stats", "error", tx.Error)
//		return result.Error
//	}
//
//	for i, _ := range memLogs {
//		memLogs[i].LogID = 0
//		_ = tx.Create(&memLogs[i])
//		// no error handling, if it fails, it fails
//	}
//
//	return nil, false
//}
//
//func createJob(tx *gorm.DB, result *gorm.DB, permJob JobPersistence) (error, bool) {
//	if errors.Is(result.Error, gorm.ErrRecordNotFound) {
//		err := tx.Create(&permJob).Error
//		if err != nil {
//			Trace("Error while creating job", "error", err)
//			return err, true
//		}
//	} else {
//		Trace("Error while creating job", "error", result.Error)
//		return result.Error, true
//	}
//	return nil, false
//}
